/**
 * Copyright (C) 2005 Gregoire Lejeune <gregoire.lejeune@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include "xslt.h"

VALUE mXML;
VALUE cXSLT;

void ruby_xslt_free( RbTxslt *pRbTxslt ) {
  if( pRbTxslt != NULL ) {
    free( pRbTxslt );
  }
}

void ruby_xslt_mark( RbTxslt *pRbTxslt ) {
  if( pRbTxslt == NULL ) return;
  if( !NIL_P(pRbTxslt->xXmlData) )        rb_gc_mark( pRbTxslt->xXmlData );
  if( !NIL_P(pRbTxslt->xXslData) )        rb_gc_mark( pRbTxslt->xXslData );
  if( !NIL_P(pRbTxslt->xXmlResultCache) ) rb_gc_mark( pRbTxslt->xXmlResultCache );
  
  if( !NIL_P(pRbTxslt->oXmlObject) )      rb_gc_mark( pRbTxslt->oXmlObject );
  if( !NIL_P(pRbTxslt->xXmlString) )      rb_gc_mark( pRbTxslt->xXmlString );
  
  if( !NIL_P(pRbTxslt->oXslObject) )      rb_gc_mark( pRbTxslt->oXslObject );
  if( !NIL_P(pRbTxslt->xXslString) )      rb_gc_mark( pRbTxslt->xXslString );
  
  if( !NIL_P(pRbTxslt->pxParams) )        rb_gc_mark( pRbTxslt->pxParams );
}

/** o = XML::XSLT.new() */
VALUE ruby_xslt_new( VALUE class ) {
  RbTxslt *pRbTxslt;

  pRbTxslt = (RbTxslt *)malloc(sizeof(RbTxslt));
  if( pRbTxslt == NULL )
    rb_raise(rb_eNoMemError, "No memory left for XSLT struct");

  pRbTxslt->iXmlType        = RUBY_XSLT_XMLSRC_TYPE_NULL;
  pRbTxslt->xXmlData        = Qnil;
  pRbTxslt->iXslType        = RUBY_XSLT_XSLSRC_TYPE_NULL;
  pRbTxslt->xXslData        = Qnil;
  pRbTxslt->iXmlResultType  = RUBY_XSLT_XMLSRC_TYPE_NULL;
  pRbTxslt->xXmlResultCache = Qnil;
  
  pRbTxslt->oXmlObject      = Qnil;
  pRbTxslt->xXmlString      = Qnil;

  pRbTxslt->oXslObject      = Qnil;
  pRbTxslt->xXslString      = Qnil;

//  pRbTxslt->pxParams        = NULL;
  pRbTxslt->pxParams        = Qnil;
  pRbTxslt->iNbParams       = 0;
  
  return( Data_Wrap_Struct( class, ruby_xslt_mark, ruby_xslt_free, pRbTxslt ) );
}

/**
 * ----------------------------------------------------------------------------
 */

VALUE ruby_xslt_xml_obj_set( VALUE self, VALUE xml_doc_obj ) {
  RbTxslt *pRbTxslt;
  Data_Get_Struct( self, RbTxslt, pRbTxslt );

  pRbTxslt->xXmlString = object_to_string( xml_doc_obj );
  if( pRbTxslt->xXmlString == Qnil ) {
    rb_raise( rb_eSystemCallError, "Can't get XML data" );
  }
  pRbTxslt->oXmlObject = xml_doc_obj;
  pRbTxslt->iXmlType   = RUBY_XSLT_XMLSRC_TYPE_STR;
  pRbTxslt->xXmlData   = pRbTxslt->xXmlString;
  
  pRbTxslt->iXmlResultType = RUBY_XSLT_XMLSRC_TYPE_NULL;

  return( Qnil );
}

VALUE ruby_xslt_xml_obj_set_d( VALUE self, VALUE xml_doc_obj ) {
  rb_warn( "XML::XSLT#xmlfile=<file> is deprecated. Please use XML::XSLT#xml=<file> !" );
  return( ruby_xslt_xml_obj_set( self, xml_doc_obj ) );
}

VALUE ruby_xslt_xml_2str_get( VALUE self ) {
  RbTxslt *pRbTxslt;
  Data_Get_Struct( self, RbTxslt, pRbTxslt );

  return( pRbTxslt->xXmlString );
}

VALUE ruby_xslt_xml_2obj_get( VALUE self ) {
  RbTxslt *pRbTxslt;
  Data_Get_Struct( self, RbTxslt, pRbTxslt );

  return( pRbTxslt->oXmlObject );
}

/**
 * ----------------------------------------------------------------------------
 */

VALUE ruby_xslt_xsl_obj_set( VALUE self, VALUE xsl_doc_obj ) {
  RbTxslt *pRbTxslt;
  Data_Get_Struct( self, RbTxslt, pRbTxslt );

  pRbTxslt->xXslString = object_to_string( xsl_doc_obj );
  if( pRbTxslt->xXslString == Qnil ) {
    rb_raise( rb_eSystemCallError, "Can't get XSL data" );
  }
  pRbTxslt->oXslObject = xsl_doc_obj;
  pRbTxslt->iXslType   = RUBY_XSLT_XSLSRC_TYPE_STR;
  pRbTxslt->xXslData   = pRbTxslt->xXslString;
  
  pRbTxslt->iXmlResultType = RUBY_XSLT_XMLSRC_TYPE_NULL;

  return( Qnil );
}

VALUE ruby_xslt_xsl_obj_set_d( VALUE self, VALUE xsl_doc_obj ) {
  rb_warning( "XML::XSLT#xslfile=<file> is deprecated. Please use XML::XSLT#xsl=<file> !" );
  return( ruby_xslt_xsl_obj_set( self, xsl_doc_obj ) );
}

VALUE ruby_xslt_xsl_2str_get( VALUE self ) {
  RbTxslt *pRbTxslt;
  Data_Get_Struct( self, RbTxslt, pRbTxslt );

  return( pRbTxslt->xXslString );
}

VALUE ruby_xslt_xsl_2obj_get( VALUE self ) {
  RbTxslt *pRbTxslt;
  Data_Get_Struct( self, RbTxslt, pRbTxslt );

  return( pRbTxslt->oXslObject );
}

/**
 * ----------------------------------------------------------------------------
 */

/** out = o.serve */
VALUE ruby_xslt_serve( VALUE self ) {
  RbTxslt *pRbTxslt;
  char *xOut;
  char **pxParams = NULL;
  
  Data_Get_Struct( self, RbTxslt, pRbTxslt );

  if( pRbTxslt->iXmlResultType == RUBY_XSLT_XMLSRC_TYPE_NULL ) {

    if( pRbTxslt->pxParams != Qnil ){
      int iCpt;
    
      pxParams = (char **)ALLOCA_N( void *, pRbTxslt->iNbParams );
      MEMZERO( pxParams, void *, pRbTxslt->iNbParams );

      for( iCpt = 0; iCpt <= pRbTxslt->iNbParams - 3; iCpt++ ) {
        pxParams[iCpt] = STR2CSTR( rb_ary_entry( pRbTxslt->pxParams, iCpt ) );
      }
    }
    
    if( pRbTxslt->iXslType != RUBY_XSLT_XSLSRC_TYPE_NULL &&
        pRbTxslt->iXmlType != RUBY_XSLT_XMLSRC_TYPE_NULL ) {
      xOut = parse( STR2CSTR( pRbTxslt->xXmlData ), pRbTxslt->iXmlType, 
                    STR2CSTR( pRbTxslt->xXslData ), pRbTxslt->iXslType,
                    pxParams );
      if( xOut == NULL ) {
        pRbTxslt->xXmlResultCache = Qnil;
        pRbTxslt->iXmlResultType = RUBY_XSLT_XMLSRC_TYPE_NULL;
      } else {
        pRbTxslt->xXmlResultCache = rb_str_new2( xOut );
        pRbTxslt->iXmlResultType = RUBY_XSLT_XMLSRC_TYPE_STR;
      }
    } else {
      pRbTxslt->xXmlResultCache = Qnil;
      pRbTxslt->iXmlResultType = RUBY_XSLT_XMLSRC_TYPE_NULL;
    }
  }

  return( pRbTxslt->xXmlResultCache );
}

/** o.save( "result.xml" ) */
VALUE ruby_xslt_save( VALUE self, VALUE xOutFilename ) {
  char *xOut;
  VALUE rOut;
  FILE *fOutFile;
  
  rOut = ruby_xslt_serve( self );
  
  if( rOut != Qnil ) {
    xOut = STR2CSTR( rOut );
  
    fOutFile = fopen( STR2CSTR( xOutFilename ), "w" );
    if( fOutFile == NULL ) {
      free( xOut );
      rb_raise( rb_eRuntimeError, "Can't create file %s\n", STR2CSTR( xOutFilename ) );
      rOut = Qnil;
    } else {
      fwrite( xOut, 1, strlen( xOut ), fOutFile );
      fclose( fOutFile );
    }
  }

  return( rOut );
}

/**
 * ----------------------------------------------------------------------------
 */

VALUE rb_xslt_errors(VALUE class) {
  return rb_cvar_get(class, rb_intern("@@errors"));
}

void xsltErrorFuncHandler(void *ctx, const char *msg, ...) {
  va_list ap;
  char *str;
  char *larger;
  int chars;
  int size = 150;
  VALUE error_arr = rb_cvar_get(cXSLT, rb_intern("@@errors"));

  /* this is taken verbatim from the libxslt python bindings */
  str = (char *) xmlMalloc(150);
  if (str == NULL)
    return;

  while (1) {
    va_start(ap, msg);
    chars = vsnprintf(str, size, msg, ap);
    va_end(ap);
    if ((chars > -1) && (chars < size))
      break;
    if (chars > -1)
      size += chars + 1;
    else
      size += 100;
    if ((larger = (char *) xmlRealloc(str, size)) == NULL) {
      xmlFree(str);
      return;
    }
    str = larger;
  }

  rb_ary_push(error_arr, rb_str_new2(str));
}
  
/**
 * ----------------------------------------------------------------------------
 */

void Init_xslt( void ) {
  mXML  = rb_define_module( "XML" );
  cXSLT = rb_define_class_under( mXML, "XSLT", rb_cObject );

  rb_define_const( cXSLT, "MAX_DEPTH",            INT2NUM(xsltMaxDepth) );
  rb_define_const( cXSLT, "MAX_SORT",             INT2NUM(XSLT_MAX_SORT) );
  rb_define_const( cXSLT, "ENGINE_VERSION",       rb_str_new2(xsltEngineVersion) );
  rb_define_const( cXSLT, "LIBXSLT_VERSION",      INT2NUM(xsltLibxsltVersion) );
  rb_define_const( cXSLT, "LIBXML_VERSION",       INT2NUM(xsltLibxmlVersion) );
  rb_define_const( cXSLT, "XSLT_NAMESPACE",       rb_str_new2(XSLT_NAMESPACE) );
  rb_define_const( cXSLT, "DEFAULT_VENDOR",       rb_str_new2(XSLT_DEFAULT_VENDOR) );
  rb_define_const( cXSLT, "DEFAULT_VERSION",      rb_str_new2(XSLT_DEFAULT_VERSION) );
  rb_define_const( cXSLT, "DEFAULT_URL",          rb_str_new2(XSLT_DEFAULT_URL) );
  rb_define_const( cXSLT, "NAMESPACE_LIBXSLT",    rb_str_new2(XSLT_LIBXSLT_NAMESPACE) );
  rb_define_const( cXSLT, "NAMESPACE_NORM_SAXON", rb_str_new2(XSLT_NORM_SAXON_NAMESPACE) );
  rb_define_const( cXSLT, "NAMESPACE_SAXON",      rb_str_new2(XSLT_SAXON_NAMESPACE) );
  rb_define_const( cXSLT, "NAMESPACE_XT",         rb_str_new2(XSLT_XT_NAMESPACE) );
  rb_define_const( cXSLT, "NAMESPACE_XALAN",      rb_str_new2(XSLT_XALAN_NAMESPACE) );

  rb_define_singleton_method( cXSLT, "new", ruby_xslt_new, 0 );
  rb_define_singleton_method( cXSLT, "extFunction", ruby_xslt_ext_function, 3);

  /* a hash of hashes. might be more elegant if it were its own class */
  /* keeps track of what functions are registered in what namespaces */
  rb_define_class_variable( cXSLT, "@@extFunctions", rb_hash_new());

  rb_define_method( cXSLT, "serve",       ruby_xslt_serve,          0 );
  rb_define_method( cXSLT, "save",        ruby_xslt_save,           1 );

  rb_define_method( cXSLT, "xml=",        ruby_xslt_xml_obj_set,    1 );
  rb_define_method( cXSLT, "xmlfile=",    ruby_xslt_xml_obj_set_d,  1 ); // DEPRECATED
  rb_define_method( cXSLT, "xml",         ruby_xslt_xml_2str_get,   0 );
  rb_define_method( cXSLT, "xmlobject",   ruby_xslt_xml_2obj_get,   0 );

  rb_define_method( cXSLT, "xsl=",        ruby_xslt_xsl_obj_set,    1 );
  rb_define_method( cXSLT, "xslfile=",    ruby_xslt_xsl_obj_set_d,  1 ); // DEPRECATED
  rb_define_method( cXSLT, "xsl",         ruby_xslt_xsl_2str_get,   0 );
  rb_define_method( cXSLT, "xslobject",   ruby_xslt_xsl_2obj_get,   0 );

  rb_define_method( cXSLT, "parameters=", ruby_xslt_parameters_set, 1 );
  rb_define_method( cXSLT, "xsl_to_s",    ruby_xslt_to_s,           0 );

  rb_define_method( cXSLT, "mediaType",   ruby_xslt_media_type,    0 );

#ifdef USE_ERROR_HANDLER
  /* this is really quite inelegant */
  xsltSetGenericErrorFunc( NULL, xsltErrorFuncHandler );
  rb_define_class_variable( cXSLT, "@@errors", rb_ary_new());
  rb_define_singleton_method( cXSLT, "errors", rb_xslt_errors, 0 );
#endif

#ifdef USE_EXSLT
  exsltRegisterAll();
#endif
}
